﻿' SQLite with RUNTIME Parameters Example (Visual Studio 2013, Visual Basic)
'    Author: Lewis Balentine, Houston, Texas, USA, October 2013
'    Website: http://www.keywild.com
'
'=======================================================================================
' PUBLIC DOMAIN:
' The author or authors of this code dedicate any and all copyright interest in this 
' code to the public domain. The author makes this dedication for the benefit of the 
' public at large and to the detriment of the author's heirs and successors. The author 
' intends this dedication to be an overt act of relinquishment in perpetuity of all 
' present and future rights to this code under copyright law.
'
'=======================================================================================
'LEGAL DISCLAIMER:
'
' The author makes no representation or warranty, either expressed or implied, 
' with respect to the data files and/or software, their quality, accuracy, or 
' fitness for any specific application. Therefore the author shall have no liability 
' to any person or any entity with respect to any liability, loss, or damage caused 
' or alleged to have been caused directly or indirectly by the use of the data files 
' and/or software. This includes, but is not limited to, interruption of service, 
' loss of data, loss of consulting or anticipatory profits or consequential damages 
' from the use of these data files and/or software.
'
' All files unless otherwise noted are the original work product of the author. 
' Unless otherwise noted these files are placed into the Public Domain for the 
' unrestricted use by anyone for any purpose. Placing these files in the public domain 
' shall in no way be construed as an obligation of the author (or his heirs and/or assigns)
' to maintain the web site, web pages, files, data or software. Further it shall in no way
' limit the author’s (or his heirs and/or assigns) options to make, produce or use versions
' of the software, data files, CAD objects or other material posted under the URL  in any 
' other commercial or non-commercial venture. 
'
' In the event of a legal dispute the court is requested to use a “reasonable person’s” 
' interpretation of the “clear intent” of this disclaimer.
'
' The use of these data files and/or software constitutes acceptance of this disclaimer.
'
'=======================================================================================
' DESCRIPTION:
' There are lots of examples of using MS's DataTable and DataGridView via the 
' Data Designer Tool. That provides an adequate solution if have your DB all
' set up and configured at Design time but it is totally useless if you want
' to select the DB at RUNTIME. I only found one example using Run Time 
' parameters (and it was riddled with errors). Thus I spent a number of hours
' figuring out a way to make it work. The primary function that you are 
' interested in is "Sql3GetDataTable". It encapsulated everything that you need
' to return a new DataTable to you application. The DataTable is similar to a
' ADO record set. It can be linked to a Data Control or data may be pulled
' directly from the data table. Both methods a demonstrated in this example 
' application.
'
' The application is a simple READ ONLY SQLite DB table browser. The SQLIte 
' ADO.NET library will allow you to edit, add or delete records but those features 
' have not been implemented in this example application.
'
' The code could be adapted to other DB engines (Access, MS SQL, Oracle, etc.) by
' changing the Library specific calls in the function "Sql3GetDataTable".
'
' The first version of this was done with Visual Basic from Visual Studio 2010
' using the SQLite.NET Class that can be downloaded from 
'        http://system.data.sqlite.org/index.html/doc/trunk/www/index.wiki
'        sqlite-netFx40-setup-bundle-x86-2010-1.0.88.0.exe
' Of five different machines there was only one that I could get the SQLite 
' ADO.Net test program to run to competition. Time for plan "B".
'
' In Visual Studio 2013 I used Nuget to install SQLite. It added: 
'     SQLite.Interop.dll
' for both X86 and X64 windows to my project.
'
' If you change your target (i.e. Net Frameworks 4.0 instead of Net Frameworks 4.5 or Net Frameworks 4.5.1
' then the only way to fix it is to completely delete your project packages directory as well as your
' packages.config. Then reload the project in Visual studio and us the NUGET package manager to 
' reinstall SQLite.Iterop.dll. This will force it to download the appropriate packages for your new target.
'
' For Net Frameworks 4.0 the packages config should say:
' <packages>
'    <package id="System.Data.SQLite" version="1.0.89.0" targetFramework="net40" />
' </packages>
'
' For each target (2.0 vs. 3.5 vs. 4.0 vs. 4.5 and  X86 vs. X64) you need at least three files:
'     System.Data.SQLite.dll
'     System.Data.SQLite.Linq.dll
'     SQLite.Interop.dll
'
' You may also get a warning: The 'packages' element is not declared.	
'     C:\Users\....\SQLiteExample(VB2013)\SQLiteExample(VB2013)\packages.config
' I have not figured out how to get rid of this.
'
' If you have more problems loading the application on the target machine then you may want to check this URL: 
'         http://stackoverflow.com/questions/13028069/unable-to-load-dll-sqlite-interop-dll
'
' ----------------------------------------------------------------------------------------------------------
'


Public Class Form1
    Public Sql3ErrorMsg As String
    Public Sql3ErrorCode As String
    Private ProcessTableName As Boolean = False
    Private Sql3FileName As String

    Public Function Sql3GetDataTable(ByVal DB As String, ByVal Sql As String) As DataTable
        ' This reference is for MS SQL but with a bit of effort it translate to SQLite.
        ' http://msdn.microsoft.com/en-us/library/vstudio/fbk67b6z(v=vs.100).aspx
        ' The biggest problem is there are numerous mistakes in the Microsoft's example code.
        ' That should be no surprise. Be sure to read the user comments.
        ' The principle difference in using SQLite and MS SQL is that you must prefix all the keywords 
        ' with "SQLite.SQLite". Yes, you are correct: It is redundant. 
        '
        ' I will not try to explain the use of the "DataAdapter", "CommandBuilder" or "DataTable" because
        ' I do not understand them. All I can tell you is that after HOURS of trial and error coding
        ' I found that this sequence works. I have tried to encapsulate it in this function. For a plain
        ' SQL command that does not produce a record set this will return an Empty DataTable (I think).
        ' For Error handling check the Public Variable: Sql3ErrorCode
        ' If it is not zero then the error message will be in: Sql3ErrorMsg
        '
        Dim DT As New DataTable()
        Dim T As String

        Sql3ErrorCode = 0
        Try
            If Dir(DB) <> "" Then
                If Sql <> "" Then
                    ' Sql = "Bad command  -- Realy bad command"         ' used in error testing
                    ' Sql = Sql & " ; garbage appended at end"          ' fails to generate an error
                    ' Sql = ""                                          ' fails to generate an error
                    ' DB = "data source=" & DB & "dddf"                 ' used in error testing
                    '
                    ' Windows uses "\\" to indicate a network drive
                    ' Apparently SQLite ADO.Net processes the file name through their Unix standard
                    ' string routines and it substitutes a single "\". This code was added to deal 
                    ' with the problem. Note that because of another bug in visual Studio 2013 I
                    ' had to use the Mid string function as the built in "Left" string function
                    ' conflicts with the control "Left" property.
                    '
                    T = Mid(DB, 1, 2)
                    If T = "\\" Then DB = "\\" & DB

                    DB = "data source=" & DB
                    Dim AdapterTemp As New SQLite.SQLiteDataAdapter(Sql, DB)
                    Dim BuilderTemp As New SQLite.SQLiteCommandBuilder()
                    BuilderTemp = New SQLite.SQLiteCommandBuilder(AdapterTemp)
                    DT.Locale = System.Globalization.CultureInfo.InvariantCulture
                    ' if there is a syntax error then the exception will be raised by this statement
                    AdapterTemp.Fill(DT)
                    BuilderTemp.Dispose()
                    AdapterTemp.Dispose()
                Else
                    Sql3ErrorMsg = "SQLite Query can not be an Empty String."
                    Sql3ErrorCode = -30
                End If
            Else
                If DB = "" Then
                    Sql3ErrorMsg = "SQLite Data file can not be an Empty String."
                    Sql3ErrorCode = -10
                Else
                    Sql3ErrorMsg = "SQLite Data file " & Chr(34) & DB & Chr(34) & " not found."
                    Sql3ErrorCode = -20
                End If
            End If
        Catch Ex As System.NotSupportedException
            Sql3ErrorMsg = Ex.Message & Chr(13) & Chr(10) & _
                "This may mean that the selected file is NOT an SQLite 3 database."
            Sql3ErrorCode = -2
        Catch Ex As System.ArgumentException
            Sql3ErrorMsg = Ex.Message & Chr(13) & Chr(10) & _
                "This may mean that the SQLite connection string is incorect."
            Sql3ErrorCode = -1
        Catch Ex As SQLite.SQLiteException
            Sql3ErrorMsg = Ex.Message
            Sql3ErrorCode = Ex.ResultCode
        End Try
        Return DT
    End Function


    Private Sub Form1_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        Dim LC As System.Drawing.Point

        ComboBox1.Enabled = False
        RecordCount.Visible = False
        TableCount.Visible = False
        Sql3FileName = ""
        LabelFileName.Text = ""
        Me.Text = "SQLite Example with RUNTIME Parameters Example (Visual Studio 2013, Visual Basic)"
        PictureBox1.Visible = False
        CellProprties.Visible = False
        LC.X = 0
        LC.Y = 0
        PictureBox2.Location = LC

    End Sub

    Private Sub Form1_Resize(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Resize
        Dim LC As System.Drawing.Point
        ' The numbers used for assition space were arrived at through intense scienticifc calculations.
        ' (OK ... I just kept trying different numbers until it looked right)
        GridOne.Width = Me.Width - (24)
        GridOne.Height = Me.Height - (GridOne.Top + 40)
        LC = CellProprties.Location
        CellProprties.Location = LC
        CellProprties.Height = GridOne.Height
        CellProprties.Width = GridOne.Width / 2
        LC.X = GridOne.Location.X + GridOne.Width / 2
        PictureBox1.Location = LC
        PictureBox1.Width = GridOne.Width / 2
        PictureBox1.Height = GridOne.Height
    End Sub

    Private Sub ComboBox1_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ComboBox1.SelectedIndexChanged
        If ProcessTableName Then
            If ComboBox1.Text <> "" Then
                ProcessTableName = False
                Call FillGridOne(Sql3FileName, ComboBox1.Text)
                ProcessTableName = True
            End If
        End If
    End Sub

    Private Sub CmdFile_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles CmdFile.Click
        DlgFile.CheckFileExists = True
        DlgFile.CheckPathExists = True
        DlgFile.FileName = ""
        DlgFile.Title = "Open SQL 3 Data Base"
        DlgFile.DefaultExt = ".db"
        DlgFile.ShowDialog()
        Sql3FileName = DlgFile.FileName
        LabelFileName.Text = "Sqlite File: " & Sql3FileName
        Call GetTableList(Sql3FileName)
    End Sub

    Private Sub GetTableList(ByVal Filename As String)
        ' For those of you coming from the VB6 and ADO enviroment this is the equivalent
        ' opening a connection and and a data set .... it is supposed to be more efficent.
        ' The "DataTable" appears to be similar to a "ADODB RecordSet"
        ' but a bit more complex, tedious and flustrating to actually use.
        ' The advantage is that unlike ADODB abd ODBC for SQLite 
        ' you can actually edit the DataTable and save it.
        '
        ' The SQL string is "similar" to the one found in the SQLite docs to return Table names.
        ' It has been modified so that it works with SQLite ADO Class Library.
        '
        Dim SqlStr As String = "Select Distinct Name from (SELECT * FROM " & _
                               "sqlite_master UNION ALL SELECT * FROM sqlite_temp_master)" & _
                               "where type = 'table' and Name=tbl_Name order by Name;"
        Dim TableTemp As New DataTable()
        TableTemp = Sql3GetDataTable(Filename, SqlStr)
        If Sql3ErrorCode <> 0 Then
            MsgBox(Sql3ErrorMsg)
        Else
            ' This could have been done by linking the ComboBox to the DataTable but
            ' I wanted to show how to pull data from the Table with out requiring a
            ' GUI Data Object on a form.
            '
            ComboBox1.Items.Clear()
            Dim R, Rows As Integer
            Dim Colname As String       ' Field Name
            Dim ColType As String       ' Field Type
            Dim ColCap As String        ' Column Caption
            Dim ColDflt As String       ' Field Default Value
            Dim ColUnq As String        ' Field Must Be Unique
            Dim ColMap As String        ' Column Mapping
            Dim ColAuto As String       ' AutoIncrements
            Rows = TableTemp.Rows.Count - 1
            TableCount.Text = "TableCount: " & Rows.ToString
            For R = 0 To Rows
                ' This is similar to the RS.MoveNext adn RS.Fields(i).Value
                ComboBox1.Items.Add(TableTemp.Rows(R).Item(0).ToString)
                ' Not needed for the operation of this function but shown as an example:
                ' To get the field Name, Type or other properties you must refer to the Columns collection
                Colname = TableTemp.Columns(0).ColumnName
                ColType = TableTemp.Columns(0).DataType.ToString
                ColCap = TableTemp.Columns(0).Caption
                ColDflt = TableTemp.Columns(0).DefaultValue.ToString
                ColUnq = TableTemp.Columns(0).Unique.ToString
                ColMap = TableTemp.Columns(0).ColumnMapping.ToString
                ColAuto = TableTemp.Columns(0).AutoIncrement.ToString
            Next
            TableTemp.Dispose()
            If ComboBox1.Items.Count > 0 Then
                ComboBox1.SelectedIndex = 0
                Call FillGridOne(Sql3FileName, ComboBox1.Text)
                ComboBox1.Enabled = True
                TableCount.Visible = True
                ProcessTableName = True
            End If
        End If
    End Sub

    Private Sub FillGridOne(ByVal FileName As String, ByVal TableName As String)
        If Dir(FileName) <> "" Then
            If TableName <> "" Then
                Dim TableOne As New DataTable()
                TableOne = Sql3GetDataTable(FileName, "Select * from " & TableName)
                BindSourceOne.DataSource = TableOne
                GridOne.DataSource = BindSourceOne
                GridOne.AutoResizeColumns()
                RecordCount.Text = "Record Count: " & TableOne.Rows.Count.ToString
                RecordCount.Visible = True
            End If
        End If
        GridOne.ScrollBars = ScrollBars.Both
        GridOne.Visible = True
        PictureBox1.Visible = False
        CellProprties.Visible = False
    End Sub

    Private Sub GridOne_CellClick(ByVal sender As Object, ByVal e As System.Windows.Forms.DataGridViewCellEventArgs) Handles GridOne.CellClick
        Dim T As String
        Dim EOL As String = Chr(13) & Chr(10)
        Dim TypeStr As String
        Dim R, C, N As Integer
        N = GridOne.Columns.Count - 1
        Dim Item As DataGridViewCell

        If e.ColumnIndex >= 0 Then
            R = e.RowIndex
            C = e.ColumnIndex
            Item = GridOne.Rows(R).Cells(C)
            T = "Press 'escape key' to close" & EOL & EOL
            T = T & "Column: " & C & EOL
            T = T & "Row: " & R & EOL
            T = T & "To String: " & Item.ToString & EOL
            T = T & "Value: " & Item.Value.ToString & EOL
            T = T & "Value Type: " & Item.ValueType.ToString & EOL
            TypeStr = Item.FormattedValue.GetType.Name.ToString
            ' For pictures this is: "Bitmap"
            T = T & "Formatted Type: " & TypeStr & EOL
            T = T & "Frozen: " & Item.Frozen.ToString & EOL
            T = T & "Edit Mode: " & Item.IsInEditMode.ToString & EOL
            T = T & "Read Only: " & Item.ReadOnly.ToString & EOL
            T = T & "State: " & Item.State.ToString & EOL
            T = T & "Tag: " & Item.Tag
            If TypeStr = "Bitmap" Then
                ' One would think that is would be easy to copy a image from the datagridview cell
                ' to a picture box and one would be wrong. Like everything else in the
                ' "new improved" versions of Visual Basic it is undocumented and somewhat less then
                ' intuitive.
                '
                ' One might expect the syntax ... but the DataGridView does not have an image property.
                ' PictureBox1.Image = Item.Image
                '
                ' This is the only obvious option.
                ' It is also the only method that was found by searching MSDN and other Visual Studio resources.
                ' It fails miserably .... 
                ' PictureBox1.Image = Item.Value

                ' You can retrieve the data but that data is not any known format and Microsoft
                ' does not bother to document the format that they are using (Microsoft's PISS POOR 
                ' DOCUMENTATION is extremely frustrating). Further it is not recognized by any internal 
                ' Visual Studio functions as being a valid graphic format.
                ' PicData = Item.Value    ' Dim PicData() as Byte
                ' My.Computer.FileSystem.WriteAllBytes("C:\tempwork\xxx.bmp", PicData, True)
                ' this gives: System.OutOfMemoryException was unhandled because it is not a valid graphic file format
                ' PictureBox1.Image = Image.FromFile("C:\tempwork\xxx.bmp") 

                ' This works. It was found by trial and many errors.

                PictureBox1.Image = Item.FormattedValue
                PictureBox1.Visible = True

            End If
            CellProprties.Text = T
            CellProprties.Enabled = True
            CellProprties.Visible = True
            CellProprties.BringToFront()
            CellProprties.Focus()
            ' This highlight the first string: "Press 'escape key' to close"
            CellProprties.Select(0, 27)
        End If
    End Sub

    Private Function BytesToImage(ByVal ImageBytes() As Byte) As Image
        Dim imgNew As Image
        Dim memImage As New System.IO.MemoryStream(ImageBytes)
        imgNew = Image.FromStream(memImage)
        Return imgNew
    End Function
    Private Sub CellProprties_KeyPress(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyPressEventArgs) Handles CellProprties.KeyPress
        If e.KeyChar = Chr(27) Then
            CellProprties.Enabled = False
            CellProprties.Visible = False
            PictureBox1.Visible = False
        End If
    End Sub

    Public Function LeftStr(DB As String, p2 As Integer) As String
        ' When I tried to use the string function Left(T,2) Microsoft's untested IDE threw an error because
        ' the "Left" string function conflicts with the "Left" Control property. It then generated this Prototype
        ' function. So I changed the name and substituted the "Mid" string function that does not generate an error.
        '
        ' 'Public Property Left As Integer' has no parameters and its return type cannot be indexed.
        '
        ' As Microsoft's systems become more complex we should learn to expect to see more of these types
        ' of errors due to their failure to implement a quality control system for their development team.
        '
        Return (Mid(DB, 1, 2))
    End Function

End Class
' End of file